var startALocalProgram = {
  /** @see #handleEvent */
  eventHandlers: {
    "load": "initialize",
    "StartLocalProgramEvent": "handleContentEvent"
  },

  /**
   * (nsIDOM)EventListener method. Routes the event to the
   * handler function, based on the event's type
   */
  handleEvent: function(e) {
    if (e.type in this.eventHandlers)
      this[this.eventHandlers[e.type]](e);
  },

  initialize: function(e) {
    // the last param indicates that we want to handle untrusted events
    // (events from content)
    document.addEventListener("StartLocalProgramEvent", this, false, true);

    var prefBranch = this.prefBranch;
    this.autoDetectPath(prefBranch, "launcherLocation");
    this.autoDetectPath(prefBranch, "programToLaunch");

    this.strings = document.getElementById("startalocalprogram-strings");
  },

  /**
   * Check if a page with the specified location should be allowed to
   * run a local program.
   * @param location DOM0 Location object.
   * @throws ReadableException if the page shouldn't get any additional permissions.
   */
  checkLocation: function(location) {
    // The pages from the following hostnames and from file:// are 
    // granted additional permissions:
    //  Class A: 10.0.0.1 / 10.255.255.255
    //  Class B: 172.16.0.1 / 172.31.255.255
    //  Class C: 192.168.0.1 / 192.168.255.255
    var ok = location.protocol == "file:";
    if (!ok) {
      var matches =
          location.hostname.match(/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/);
      ok = matches && // pages accessed via domain names won't get special privileges
          (
           (matches[1] == "10") || // Class A
           (matches[1] == "172" && // Class B
            parseInt(matches[2], 10) >= 16 &&
            parseInt(matches[2], 10) < 32) ||
           (matches[1] == "192" && matches[2] == "168")
          ); // Class C
    }
    if (!ok) {
      throw new this.ReadableException(
        this.strings.getFormattedString("accessNotAllowed",
                                        [location.hostname]));
    }
  },

  /**
   * This function gets called when a webpage requests a local program
   * to be launched. The event's target is a node with DOM attributes
   * param0, param1, ..., paramN specifying the command line parameters
   * to pass to the application.
   * The application to launch is specified in the extension's preferences.
   *
   * If the function handles the event, it lets the caller know the
   * event was handled by setting the "done" attribute on the element.
   *
   * In case there was an error, the function sets the attribute
   * "errorMessage" to a string describing the failure, which can be
   * displayed to the user.
   * In case of success, the function removes an existing "errorMessage"
   * attribute if there is one.
   */
  handleContentEvent: function(ev) {
    var data = ev.target;
    try {
      // let the web page know the event was handled
      data.setAttribute("done", "1");

      // throws ReadableException if access should be denied
      this.checkLocation(data.ownerDocument.location);

      // obtain the parameters from the web page
      var params = [];
      for (var i=0; data.hasAttribute("param"+i); i++) {
        params.push(data.getAttribute("param"+i));
      }

      this.launchProgram(params);

      data.removeAttribute("errorMessage");
    } catch(e if e instanceof this.ReadableException) {
      // This is an exception with message suitable for showing to the user.
      data.setAttribute("errorMessage", e.message);
    } catch(e) {
      // this is not supposed to happen
      Components.utils.reportError(e);
      data.setAttribute("errorMessage",
          this.strings.getString("unexpectedError"));
    }
  },

  get prefBranch() {
    return Components.classes["@mozilla.org/preferences-service;1"]
                      .getService(Components.interfaces.nsIPrefService)
                      .getBranch("extensions.startalocalprogram.");
  },

  /**
   * Runs the program specified in extension's preferences.
   * @param params an array of parameters to pass to the program.
   */
  launchProgram: function(params) {
    var prefBranch = this.prefBranch;

    var args = params;

    var launcherLocation = this.getFilePref(prefBranch, "launcherLocation");
    var programToLaunch = this.getFilePref(prefBranch, "programToLaunch");

    if (!programToLaunch) {
      throw new this.ReadableException(
          this.strings.getString("programNotSpecified"));
    }

    if (launcherLocation) {
      // we'll run the program specified in |launcherLocation| with the
      // first param being the path to the actual program.
      args.unshift(programToLaunch.path);
    } else {
      // just run the program
      launcherLocation = programToLaunch;
    }

    try {
      var process = Components.classes["@mozilla.org/process/util;1"]
                              .createInstance(Components.interfaces.nsIProcess);
      process.init(launcherLocation);
      process.run(false, args, args.length);
    } catch(e) {
      Components.utils.reportError(e);
      throw new this.ReadableException(
        this.strings.getFormattedString("couldNotRunProgram",
                                        [launcherLocation.path]));
    }
  },

  /**
   * Returns the value of the specified pref as an nsIFile. Calls
   * autoDetectPath if necessary.
   *
   * @param prefBranch {nsIPrefBranch} to get the file from.
   * @param prefName {string} the name of unichar preference to get,
   *                          relative to the prefBranch branch.
   * @returns an nsIFile stored in the specified pref, may be null.
   * @throws ReadableException if the preference is not set
   *         or has a wrong value.
   */
  getFilePref: function(prefBranch, prefName) {
    try {
      var path = prefBranch.getComplexValue(prefName,
          Components.interfaces.nsISupportsString).data;
    } catch(e) {
      Components.utils.reportError(e);
      throw new this.ReadableException(
        this.strings.getString("couldNotReadPrefs"));
    }

    if (path == "") {
      return null;
    }
    if (path == "autodetect") {
      // a special value meaning that the extension should use the
      // default OS-specific value for the field.
      path = this.autoDetectPath(prefBranch, prefName);
    }

    var file = Components.classes["@mozilla.org/file/local;1"]
                         .createInstance(Components.interfaces.nsILocalFile);
    try {
      file.initWithPath(path); // throws if the path is malformed
    } catch(e) {
      Components.utils.reportError(e);
      throw new this.ReadableException(
        this.strings.getFormattedString("badPath", [path]));
    }

    if (!file.exists()) {
      throw new this.ReadableException(
        this.strings.getFormattedString("fileNotFound", [path]));
    }
    return file;
  },

  /**
   * Determines the default path to use for the specified pref
   * and updates the pref's value.
   * @returns the detected path.
   */
  autoDetectPath: function(prefBranch, prefName) {
    var os = "unknown";
    if (navigator.platform.search(/win32/i) > -1) {
      os = "windows";
    } else if (navigator.platform.search(/linux/i) > -1) {
      os = "linux";
    } else if (navigator.platform.search(/mac/i) > -1) {
      os = "mac";
    }

    if (prefName == "launcherLocation" ||
        prefName == "programToLaunch")
    {
      var detectedValue = prefBranch.getComplexValue(
        prefName + "." + os, Components.interfaces.nsISupportsString).data;
    } else {
      throw "Whoops. Unexpected prefName passed to autoDetectPath: " + prefName;
    }

    var str = Components.classes["@mozilla.org/supports-string;1"]
                        .createInstance(Components.interfaces.nsISupportsString);
    str.data = detectedValue;
    prefBranch.setComplexValue(prefName,
        Components.interfaces.nsISupportsString, str);
    return detectedValue;
   },

  /**
   * An exception, whose message is suitable for showing to the user.
   * @constructor
   */
  ReadableException: function(message) {
    this.message = message;
  }
};

window.addEventListener("load", startALocalProgram, false);
